
-- Why are so many list functions written recursively?
-- Because the list data type is defined recursively.

-- Recursive functions defined on lists often have the following pattern.

-- f :: [a] -> type of result
-- f [] = result for empty list
-- f (x:xs) = result defined using x and (f xs)

-- The first equation stops the recursion by defining a result for
-- the empty list. The second equation uses recursion and it shows
-- how to build the result using the first item from the list and
-- the recursive result created by f from the rest of the list.


-- Here are several examples of recursive functions defined on lists.


-- here, the pattern h:t is read as "the head and tail of a list"
size [] = 0
size (h:t) = 1 + size t

-- sum the elements of a list
mySum [] = 0
mySum (h:t) = h + mySum t

{- Java versions of sum

   // sum a linked list recursively
   public static int sumLR(Node head)
   {
      if (null == head) // empty list
         return 0;
      int r = sumLR(head.tail); // recursion
      return head.data + r;
   }

   // sum a linked list iteratively
   public static int sumLI(Node head)
   {
      int result = 0;
      while (null != head)
      {
         result += head.data;  // mutate result variable
         head = head.tail;     // mutate list variable
      }
      return result;
   }

   // sum an array recursively
   public static int sumAR(int[] a, int low, int hi)
   {
      if (null == a) { /* ???? */ }  // what to do for a "null" array?   
      if (low > hi)                  // an empty array
         return 0;
      int r = sumAR(a, low+1, hi);   // recursion
      return a[low] + r;
   }

   // sum an array iteratively
   public static int sumAI(int[] a)
   {
      if (null == a) { /* ???? */ }       // what to do for a "null" array?
      int result = 0;
      for (int i = 0; i < a.length; i++)  // works for an empty array
      {
         result += a[i]; // mutate the result variable
      }
      return result;
   }

-}


-- here, the pattern x:xs is read as "a list with x and more x's"
prod [] = 1
prod (x:xs) = x * prod xs


-- Find the maximum or minimum element of a non-empty list.
-- maxList [] = 
maxList (x:[]) = x
maxList (x:xs) = max x (maxList xs)

minList (x:[]) = x
minList (x:xs) = min x (minList xs)


maxList' (x:[]) = x
maxList' (x:xs) | (x > max) = x
                | otherwise = max
         where max = maxList' xs



-- Here is a pair of "mutually recursive" functions.
-- Neither function is recursive (neither function 
-- calls itself directly), but each function can call
-- itself indirectly, through a call to the other one.
even' n | n < 0 = even' (-n)
even' 0 = True
even' n = odd' (n-1)
odd'  n | n < 0 = odd' (-n)
odd'  0 = False
odd'  1 = True
odd'  n = even' (n-1)



-- Here is a function that mimics Haskell's builtin range operator.
range' start step = start:(range' (start + step) step)



-- remove the odd numbers from a list of integers
allEven1 [] = []
allEven1 (h:t) = if even h then 
                    h:allEven1 t 
                 else 
                    allEven1 t

-- same function defined using a list comprehension
allEven2 x = [y | y <- x, even y]



-- Here is a generalization of allEven1.
-- p stands for "predicate"
-- Be sure to check the type of this function.
myFilter1 p (h:t) = if (p h) then 
                       h:myFilter1 p t 
                    else 
                       myFilter1 p t

-- same function defined using a list comprehension
myFilter2 p x = [y | y <- x, p y]
